package pds.label;

import java.io.BufferedReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.lang.Character;
/**
 * PDSElement is a class that contains a single definition or line 
 * as specified in the PDS Object Defnition Language (ODL). 
 * In this context a line may be an simple line of text, a block of 
 * commented text, or a keyword/value pair. A value may extend 
 * over more than one physical line if it is quoted or part of 
 * a value set.
 *
 * @author      Todd King
 * @author      Planetary Data System
 * @version     1.0, 02/21/03
 * @since       1.0
 */
public class PDSElement {
 	// Enumeration of possible value type
 	/** Unspecified grouping type. */	public static final int	TYPE_NONE		= 0;
 	/** A blank line */					public static final int	TYPE_BLANK_LINE	= 1;
 	/** A comment */					public static final int	TYPE_COMMENT	= 2;
 	
 	/** The text found before the equal sign of an element. If the element does not
 	 * contain and equal sign the value is blank.
 	 */
 	public String		mKeyword = "";
 	
 	/** An array containing a PDSValue object for each value following
 	 * 						the equal sign in the element. There may be any number of
 	 *						values.
 	 */
 	public PDSValue	mValue = new PDSValue();
 	
 	/** The basic type of the value. Possible values include:
 	 *						<dt>TYPE_NONE</dt><dd>Unspecified grouping type</dd>
 	 *						<dt>TYPE_ORDERED</dt><dd>An ordered list of values</dd>
 	 *						<dt>TYPE_UNORDERED</dt><dd>A unordered list of values</dd>
 	 */
 	public int			mType = TYPE_NONE;
 	
 	/** The comment text found within the element. If the element does not contain a comment the vlaue is blank.
 	 */
 	public String		mComment = "";
 	
 	/** The raw line as read from the file.
 	 */
 	public char[]		mRaw = null;
 	
 	/** Maximum line length when printing. Really its 80 including the DOS style new line (2 bytes). 
 	 * All lines are wrapped if they exceed this length
 	 */
 	public int					mMaxLength = 78;	
 	
 	/** The count of the number of physical lines parsed into this element
 	 */
 	 public int				mLineCount = 0;
 	 
 	/** Indicates whether a symtax error occurred will parsing or reading a line.
 	 */
 	 public boolean			mSyntaxError = false;
 	 
 	/** Creates an instance of a PDSElement */
 	public PDSElement() {
 	}
 	
 	/** Creates an instance of a PDSElement */
 	public PDSElement(int line) {
 		mLineCount = line;
 	}
 	
    /** 
     * Parses the next element from a file stream. An element
	 * is a single definition or line as specified in the PDS 
	 * Object Defnition Language (ODL). In this context a line 
	 * may be an simple line of text, a block of commented text, 
	 * or a keyword/value pair. A value may extend over more than 
	 * one physical line if it is quoted or part of a value set.
	 * 
     * @param reader    the input file stream.
     *
     * @return          <code>true</code> if an element was parsed from the stream;
     *                  <code>false</code> if the end of file or an error was encountered.
     * @see             FileInputStream
     * @since           1.0
     */
     // Note: We use character arrays for optimization. Long strings
     //       are very slow to parse.
	public boolean parse(BufferedReader reader) 
		throws PDSException
	{
		char		lastC = 0;
		char		c = 0;
		int			quote = ' ';
		int			i;
		int 		total = 0;
		boolean		inQuote = false;
		boolean		inBlock = false;
		boolean		inComment = false;
		boolean		inUnits = false;
		boolean		add = true;
		int			commentStart;
		int			commentEnd;
		boolean		scanning = true;
		String		buffer;
		
		try {
			while(scanning) {
				mRaw = readLine(reader);
				if(mRaw == null) return false;		// End of file
				
				total = strlen(mRaw);
				// if(total == 0) continue;	// Keep reading
						
				// Now parse the raw string
				// Extract comment from string
				commentStart = 0;
				commentEnd = 0;
				inComment = false;
				inQuote = false;
				lastC = 0;
				for(i = 0; i < total; i++) {
					c = mRaw[i];
					if(!inComment) {	// Check if in quote
						if(c == '"') { if(inQuote) inQuote = false; else inQuote = true; }
					}
					if(!inQuote) {	// Check if in comment
						if(lastC == '/' && c == '*') { commentStart = i-1; inComment = true; }
						if(lastC == '*' && c == '/') { commentEnd = i; inComment = false; add = false; }
					}
					lastC = c;
					
					if(inComment) { mComment += (char) c; continue; }
				}
				if(mComment.length() > 0) mComment = mComment.substring(1);	// Remove leading "*";
				if(mComment.length() > 0) mComment = mComment.substring(0, mComment.length() - 1);	// Remove trailing "*"
				
				// Remove comment from raw buffer;
				if(commentStart != commentEnd) {
					deleteString(mRaw, commentStart, commentEnd);
				}
				
				trimSpace(mRaw);
				total = strlen(mRaw);
				if(total == 0) {
					if(commentEnd > 0) {
						mType = TYPE_COMMENT;
					} else {
						mType = TYPE_BLANK_LINE;
					}
					return true;
				}
				
				scanning = false;
			}

			// Check for END statement
			if(total >= 3) {
				buffer = new String(mRaw, 0, 3);
				if(total > 3) c = mRaw[3];
				else c = ' ';
				if(buffer.compareTo("END") == 0 && Character.isWhitespace(c)) {
					mKeyword = "END";
					return false;
				}
			}
			
			// Extract keyword and parse value
			i = indexOf(mRaw, '=');
			if(i == -1) { // Just a word - check if "END"
				mKeyword = new String(mRaw, 0, total); 
				if(total > 14) {	// Check for CCSD tag
					if(mKeyword.substring(0, 14).compareTo("CCSD3ZF0000100") == 0) return true;
				}
				if(mKeyword.compareTo("END") == 0) return false; // End of label - no more keywords
				throw(new PDSException("Syntax error. Unrecognized plain word (" + mKeyword + ") at line " + mLineCount));
			}
		
			// Extract keyword	
			mKeyword = new String(mRaw, 0, i);
			mKeyword = mKeyword.trim();
	
			// Extract value
			return mValue.parse(mRaw, i+1);
		} catch(PDSException e) {
			throw(e);
		}

	}

	public int indexOf(char[] cbuff, char c)
	{
		for(int i = 0; i < cbuff.length; i++) { if(cbuff[i] == c) return i; }
		return -1;
	}
	
	public void deleteString(char[] cbuff, int start, int end) 
	{
		int	i;
		int	n = start;
		
		for(i = end+1; i < cbuff.length; i++) { cbuff[n++] = cbuff[i]; }
		for(i = n; i < cbuff.length; i++) { cbuff[i] = '\0'; }	
	}
	
	public int strlen(char[] cbuff)
	{
		return strlen(cbuff, 0);
	}
	
	public int strlen(char[] cbuff, int startAt)
	{
		int	total = cbuff.length;
		for(int i = startAt; i < total; i++) {if(cbuff[i] == '\0') return i-startAt; }
		return cbuff.length-startAt;
	}
	
    /** 
     * Read the next element definition from an input file stream.
	 * An element may extend over more than one physical line if it 
	 * is quoted or part of a value set.
	 * 
     * @param reader      the input file stream.
     *
     * @return          <code>true</code> if an element was parsed from the stream;
     *                  <code>false</code> if the end of file or an error was encountered.
     *
     * @see             FileInputStream
     * @since           1.0
     */
	public char[] readLine(BufferedReader reader) 
		throws PDSException
	{
		int			lastC	= 0;
		int			c		= 0;
		int			quote = ' ';
		boolean		inQuote = false;
		boolean		inBlock = false;
		boolean		inComment = false;
		boolean		inUnits = false;
		boolean		inValue = false;
		boolean		foundValue = false;
		
		int			n 			= 0;
		int			total		= 0;

		int			maxBuffer	= 4048;
		char[]		buffer = new char[maxBuffer];
		ArrayList	buffList = new ArrayList();

		mSyntaxError = false;
		try {
			while((c = reader.read()) != -1) {
				buffer[n] = (char) c;
				n++;
				total++;
				if(n == maxBuffer) {
					buffList.add(buffer);
					buffer = new char[4048];
					n = 0;
				}
				if(c == '\n') mLineCount++;
				if(!inComment) {	// Check for quoting
					if(c == '"') { if(inQuote) inQuote = false; else inQuote = true; }
				}
				if(!inQuote) {	// Check if in comment
					if(lastC == '/' && c == '*') inComment = true;
					if(lastC == '*' && c == '/') inComment = false;
				}
				lastC = c;
				if(inValue && !isWhitespace(c)) { foundValue = true; }
				if(inQuote) continue;
				if(inComment) continue;
				
				// if(c == '(' || c == '{') { if(inBlock) { mSyntaxError = true; return null; } inBlock = true; }
				// if(c == ')' || c == '}') { if(!inBlock) { mSyntaxError = true; return null; } inBlock = false; }
				
				if(c == '\'') { if(inBlock) inBlock = false; else inBlock = true; }
				if(c == '<') { if(inUnits) { mSyntaxError = true; return null; } inUnits = true; }
				if(c == '>') { if(!inUnits) { mSyntaxError = true; return null; } inUnits = false; }
				if(c == '=') { inValue = true; }
				
				if(c == '\n' && foundValue && !inQuote && !inBlock && !inUnits) break;
			}
		} catch(Exception e) {
			buffer = null;
			throw(new PDSException(e.getMessage()) );
		}
		
		if(c == -1 && total == 0) return null;	// End of file
		if(n > 0) {
			buffList.add(buffer);
		}
		
		// Build up return string
		int	i;
		int	start = 0;
		int	end;
		int	remaining = total;
		char[] cbuff = new char[total];
		Iterator t = buffList.iterator();
		while(t.hasNext()) {
			buffer = (char[]) t.next();
			n = remaining;
			if(n > maxBuffer) n = maxBuffer;
			for(i = 0; i < n; i++) cbuff[start+i] = buffer[i];
			start += n;
			remaining -= n;
		}
		
		return cbuff;
	}
	
	
    /** 
     * Trims the leading and trailing white space from a character array.
     * White space characters are tab, space, new line, return.
	 * 
     * @param cbuff    the character array to trim.
     *
     * @return          <code>true</code> if the character is white space;
     *                  <code>false</code> otherwise.
     *
     * @since           1.0
     */
	public void trimSpace(char[] cbuff) {
		trimSpace(cbuff, 0);
	}
		
    /** 
     * Trims the leading and trailing white space from a character array.
     * White space characters are tab, space, new line, return.
	 * 
     * @param cbuff    the character array to trim.
     *
     * @return          <code>true</code> if the character is white space;
     *                  <code>false</code> otherwise.
     *
     * @since           1.0
     */
	public void trimSpace(char[] cbuff, int startAt) {
		// Trim leading and trailing space
		int	start;
		int	end;
		int	i, n;
		int total = cbuff.length;
		
		for(end = total-1; end > -1; end--) {
			if(! Character.isWhitespace(cbuff[end])) { break; }
		}
		for(start = startAt; start < end; start++) {
			if(! Character.isWhitespace(cbuff[start])) { break; }
		}
		
		if(start > startAt || end < total-1) {
			n = startAt;
			for(i = start; i <= end; i++) cbuff[n++] = cbuff[i];
			for(i = end+1; i < total; i++) cbuff[n++] = '\0';	// Zero fill
		}
		return;
	}
		
    /** 
     * Determines if a character is a white space character.
     * White space characters are tab, space, new line, return.
	 * 
     * @param c    the character to check.
     *
     * @return          <code>true</code> if the character is white space;
     *                  <code>false</code> otherwise.
     *
     * @since           1.0
     */
	public boolean isWhitespace(int c) {
		if(c == ' ') return true;
		if(c == '\n') return true;
		if(c == '\r') return true;
		if(c == '\t') return true;
		
		return false;
	}
	

	
    /** 
     * Create a copy of the element and return a new instance
     * of a PDSElement. 
	 * 
     * @return          a new instance of a PDSElement with the same content as this element.
     * @since           1.0
     */
	public PDSElement copy() {
		PDSElement	element = new PDSElement();
		
		element.mComment = mComment;
		element.mKeyword = mKeyword;
		element.mRaw = mRaw;
		element.mType = mType;
		element.mValue = mValue;
		
		return element;
	}

    /** 
     * Print the element according to PDS specifications for label
     * files to Syste.out. The line that is output can be indented with the equal sign
     * (when present) placed at a fixed position. The line is printed to System.out.
	 * 
     * @param indent    the number of spaces to indent for each level.
     * @param equal		the number of spaces from the end of the indent
     *					to align the equal sign for elements which have 
     *					a keyword and value.
     * @param level		the current level of indenting. The number of spaces
     *					the element will be indented is level*indent
     * @since           1.0
     */
	public void print(int indent, int equal, int level) {
		print(System.out, indent, equal, level);
	}
	
    /** 
     * Print the element according to PDS specifications for label
     * files. The line that is output can be indented with the equal sign
     * (when present) placed at a fixed position. The line is printed to System.out.
	 * 
     * @param out    	the stream to print the element to.
     * @param indent    the number of spaces to indent for each level.
     * @param equal		the number of spaces from the end of the indent
     *					to align the equal sign for elements which have 
     *					a keyword and value.
     * @param level		the current level of indenting. The number of spaces
     *					the element will be indented is level*indent
     * @since           1.0
     */
	public void print(PrintStream out, int indent, int equal, int level) {
		int			i, k, n;
		int			realIndent = 0;
		PDSValue	value;
		int			col;
		boolean		checkWrap;
		String		newline = "\r\n";
		
		switch(mType) {
		case TYPE_BLANK_LINE:
			out.print(newline);
			return;
		case TYPE_COMMENT:
			out.print(mComment + newline);
			return;
		}
		
		// Print keyword
		if(mKeyword.length() != 0) {
			realIndent = indent * level;
			printSpaces(out, realIndent);
			out.print(mKeyword);
		}
		
				
		// Print value
		mValue.print(out, indent, equal, level);
		
		// Print comment
		if(mComment.length() > 0) {
			if(mKeyword.length() > 0) out.print(" ");
			out.print("/*" + mComment + "*/");
		}
		
		out.print(newline);
	} 

    /** 
     * Dump all information about an element.
 	 * 
     * @param out    	the stream to print the element to.
     *
     * @since           1.0
     */
	public void dump(PrintStream out) {
		PDSValue	value;
		
		out.println("Keyword: " + mKeyword);
		out.println("Line Count: " + mLineCount);
		out.println("Type: " + mType);
		out.println("Comment: " + mComment);
		out.println("Raw line: " + mRaw);
		out.println("--- Parsed values ---");
		mValue.dump(out);
	}
	
    /** 
     * Print a string of spaces to an output stream. 
 	 * 
     * @param out    	the stream to print the element to.
     * @param count     the number of spaces to print.
     *
     * @since           1.0
     */
	public void printSpaces(PrintStream out, int count) {
		for(int i = 0; i < count; i++) out.print(" ");
	}
	
}
	
 